/*
 * Copyright 2007-2009, Thomas Maier-Komor
 * Licenses: LGPL, see file lgpl.txt
 */

#ifndef _REENTRANT
#define _REENTRANT
#endif
#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <link.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <utime.h>


typedef enum {
	do_close	/*close*/,
	do_fclose	/*fclose*/,
	do_write	/*write*/,
	do_fsync	/*fsync*/,
	do_fflush	/*fflush*/,
	do_terminate	/*terminate*/,
	do_futimesat	/*futimesat*/,
	do_utimes	/*utimes*/,
	do_utime	/*utime*/,
	do_chmod	/*chmod*/
} task_t;

typedef struct op
{
	task_t op;
	void *buf;
	size_t s,bs;
	struct op *next;
	union {
		struct timeval times[2];
		struct utimbuf utime;
	} time;
} op_t;

typedef struct operator
{
	op_t *volatile head, *volatile tail, *volatile unused;
	FILE *str;
	struct operator *next;
	volatile int fd, wr_err, at, idle;
	volatile long mem;
	char fn[PATH_MAX];
	pthread_t thread;
	pthread_mutex_t lock;
	pthread_cond_t oom;
	pthread_cond_t start;
} operator_t;

typedef struct fd2name
{
	int fd;
	char name[PATH_MAX];
	struct fd2name *next;
} fd2name_t;

typedef struct memlog
{
	size_t avail;
	char *buf, *at;
	struct memlog *next;
} memlog_t;
	


static fd2name_t *fd2name;
static operator_t *Operators = 0;
static void *libc_hdl = 0;
static int Verbose = 1, Priority = 10, MaxThreads = 32, NumThreads = 0,
	   NumDirect = 0, LogFile = STDERR_FILENO, Abort = 0;
static size_t MemPerThread = 16 << 20;
static memlog_t *LogHead = 0, *LogTail = 0;
static pthread_mutex_t LogMtx = PTHREAD_MUTEX_INITIALIZER;
static sem_t Idle;
static int (*realclose)(int) = 0,
	   (*realfsync)(int) = 0,
	   (*realfchmod)(int,mode_t) = 0,
	   (*realfflush)(FILE *) = 0,
	   (*realwrite)(int, const void *, size_t) = 0,
	   (*realread)(int, void *, size_t) = 0,
	   (*realopen)(const char *, int, ...) = 0,
	   (*realopen64)(const char *, int, ...) = 0,
	   (*realopenat)(int, const char *, int, ...) = 0,
	   (*realopenat64)(int, const char *, int, ...) = 0,
	   (*realfclose)(FILE *) = 0,
	   (*realcreat)(const char *, mode_t) = 0,
	   (*realcreat64)(const char *, mode_t) = 0,
	   (*realutime)(const char *path, const struct utimbuf *times),
	   (*realutimes)(const char *path, const struct timeval times[2]),
	   (*realfutimesat)(int,const char *path, const struct timeval times[2]),
	   (*realchmod)(const char *path, mode_t mode);
static FILE
	   *(*realfdopen)(int,const char *) = 0,
	   *(*realfopen)(const char *, const char *) = 0,
	   *(*realfopen64)(const char *, const char *) = 0;
static void (*realexit)(int) = 0, (*real_exit)(int) = 0, (*realabort)(void) = 0;
static void *writer_thread(void *);
static void error(const char *str, ...);
static void debug(const char *str, ...);
static void logmsg(const char *str, ...);



void init(void)
{
	const char *V,*T,*P,*M,*L,*em = 0,*F;
	int err;
	size_t sem_max;

	if (libc_hdl != 0)
		return;
	assert(sizeof(mode_t) == sizeof(size_t));
#ifdef __linux
	libc_hdl = dlopen("libc.so.6",RTLD_NOW|RTLD_LOCAL);
#else
	libc_hdl = dlopen("libc.so",RTLD_NOW|RTLD_LOCAL);
#endif
	assert(libc_hdl);
	realopen64 = (int (*)(const char *,int,...)) dlsym(libc_hdl,"open64");
	assert(realopen64);
	realopen = (int (*)(const char *,int,...)) dlsym(libc_hdl,"open");
	assert(realopen);
	realfdopen = (FILE *(*)(int,const char *)) dlsym(libc_hdl,"fdopen");
	assert(realfdopen);
	realfopen64 = (FILE *(*)(const char *,const char *)) dlsym(libc_hdl,"fopen64");
	assert(realfopen64);
	realfopen = (FILE *(*)(const char *,const char *)) dlsym(libc_hdl,"fopen");
	assert(realfopen);
	realfclose = (int (*)(FILE *)) dlsym(libc_hdl,"fclose");
	assert(realfclose);
	realcreat64 = (int (*)(const char *,mode_t)) dlsym(libc_hdl,"creat64");
	assert(realcreat64);
	realutime = (int (*)(const char *path, const struct utimbuf *times)) dlsym(libc_hdl,"utime");
	assert(realutime);
	realutimes = (int (*)(const char *path, const struct timeval times[2])) dlsym(libc_hdl,"utimes");
	assert(realutimes);
	realchmod = (int (*)(const char *path, mode_t mode)) dlsym(libc_hdl,"chmod");
	assert(realchmod);
	realcreat = (int (*)(const char *,mode_t)) dlsym(libc_hdl,"creat");
	assert(realcreat);
#ifdef __sun
	realfutimesat = (int (*)(int,const char *path, const struct timeval times[2])) dlsym(libc_hdl,"futimesat");
	assert(realfutimesat);
	realopenat64 = (int (*)(int,const char *,int,...)) dlsym(libc_hdl,"openat64");
	assert(realopenat64);
	realopenat = (int (*)(int,const char *,int,...)) dlsym(libc_hdl,"openat");
	assert(realopenat);
#endif
	realclose = (int (*)(int)) dlsym(libc_hdl,"close");
	assert(realclose);
	realfsync = (int (*)(int)) dlsym(libc_hdl,"fsync");
	assert(realfsync);
	realfchmod = (int (*)(int,mode_t)) dlsym(libc_hdl,"fchmod");
	assert(realfchmod);
	realfflush = (int (*)(FILE *)) dlsym(libc_hdl,"fflush");
	assert(realfflush);
	realread = (int (*)(int, void *, size_t)) dlsym(libc_hdl,"read");
	assert(realread);
	realwrite = (int (*)(int, const void *, size_t)) dlsym(libc_hdl,"write");
	assert(realwrite);
	realexit = (void (*)(int)) dlsym(libc_hdl,"exit");
	assert(realexit);
	real_exit = (void (*)(int)) dlsym(libc_hdl,"_exit");
	assert(real_exit);
	realabort = (void (*)(void)) dlsym(libc_hdl,"abort");
	assert(realabort);
	if (V = getenv("MTWRITE_VERBOSE"))
		Verbose = atol(V);
	if (L = getenv("MTWRITE_MEMLOG")) {
		if (strcmp(L,"0")) {
			LogHead = LogTail = calloc(1,sizeof(memlog_t));
			assert(LogHead);
			LogHead->buf = LogHead->at = malloc(4<<20);
			LogHead->avail = 4<<20;
			logmsg("memlogging enabled\n");
		}
	}
	if (F = getenv("MTWRITE_LOGFILE")) {
		if (strcmp(F,"")) {
			LogFile = realopen64(F,O_CREAT|O_TRUNC|O_APPEND|O_WRONLY,0666);
			if (LogFile == -1) {
				em = strerror(errno);
				LogFile = STDERR_FILENO;
			}
		}
	}
	if (Verbose) {
		logmsg("MT-write version r90214 - Copyright 2007-2009, Thomas Maier-Komor\n");
		logmsg("Licenses: LGPL, see file lgpl.txt for details\n");
	}
	if (F) {
		if (em)
			error("could not set logmsg file to %s: %s\n",F,em);
		else
			logmsg("log file set to %s\n",F);
	}
	if (T = getenv("MTWRITE_MAXTHREADS")) {
		MaxThreads = atol(T);
		if (MaxThreads < 0) {
			error("MTWRITE_MAXTHREADS must be a positive integer - resetting to defaults.");
			MaxThreads = 16;
		}
		if (MaxThreads > 0)
			logmsg("Maximum number of wroker threads: %d\n",MaxThreads);
		else
			logmsg("Maximum number of wroker threads: unlimited\n");
	}
	if (P = getenv("MTWRITE_PRIORITY"))
		Priority = atol(P);
	if (M = getenv("MTWRITE_MEMPERTHREAD")) {
		char dim,s;
		switch (s = sscanf(M,"%ld%c",&MemPerThread,&dim)) {
		default:
			error("Invalid MTWRITE_MEMPERTHREAD specification ignored.\n");
			MemPerThread = 16 << 20;
			break;
		case 2:
			switch (dim) {
			case 'K':
			case 'k':
				MemPerThread <<= 10;
				break;
			case 'M':
				MemPerThread <<= 20;
				break;
			case 'G':
				MemPerThread <<= 30;
				break;
			default:
				error("Invalid MTWRITE_MEMPERTHREAD specification ignored.\n");
				break;
			}
			/*FALLTHRU*/
		case 1:
			if (MemPerThread)
				logmsg("Maximum memory per thread: %ld\n",MemPerThread);
			else
				logmsg("Maximum memory per thread: unlimited\n");
			break;
		}
	}
	sem_max = sysconf(_SC_SEM_VALUE_MAX);
	assert(sem_max > 0);
	err = sem_init(&Idle,0,MaxThreads ? MaxThreads : sem_max);
	assert(err != -1);
	debug("init successful (Verbose = %d,MaxThreads = %d, MemPerThread = %lld)\n",
			Verbose,MaxThreads,MemPerThread);
}


#ifdef __sun
#define init()

static void add_file(int fd, const char *name)
{
	fd2name_t *n = fd2name;
	if (fd < 0)
		return;
	while (n) {
		if (n->fd == fd)
			break;
		n = n->next;
	}
	if (n == 0) {
		n = malloc(sizeof(fd2name_t));
		n->fd = fd;
		n->next = fd2name;
		fd2name = n;
	}
	(void) strcpy(n->name,name);
}


static const char *get_name(int fd)
{
	if (fd > 0) {
		fd2name_t *n = fd2name;
		while (n) {
			if (n->fd == fd)
				return n->name;
			n = n->next;
		}
		realabort();
		return 0;
	} else {
		assert(fd == AT_FDCWD);
		return ".";
	}

}
#else
#define add_file(a,b)
#endif


static inline void writelog(const char *str, va_list val)
{

	if (LogTail == 0) {
		char msg[256] = "mtwrite.so: ";
		size_t l = vsnprintf(msg+12,sizeof(msg)-12,str,val);
		realwrite(LogFile, msg, l + 12);
	} else {
		size_t ml = vsnprintf(0,0,str,val);
		int err = pthread_mutex_lock(&LogMtx);
		assert(err == 0);
		if (LogTail->avail < ml) {
			LogTail->next = calloc(1,sizeof(memlog_t));
			LogTail = LogTail->next;
			LogTail->at = LogTail->buf = malloc(4 << 20);
			LogTail->avail = 4 << 20;
		}
		(void) vsprintf(LogTail->at,str,val);
		LogTail->at += ml;
		LogTail->avail -= ml;
		err = pthread_mutex_unlock(&LogMtx);
		assert(err == 0);
	}
}


static void error(const char *str, ...)
{
	va_list val;
	va_start(val,str);
	writelog(str, val);
	va_end(val);
}


static void logmsg(const char *str, ...)
{
	if (Verbose >= 1) {
		va_list val;
		va_start(val,str);
		writelog(str, val);
		va_end(val);
	}
}


static void debug(const char *str, ...)
{
	if (Verbose >= 2) {
		va_list val;
		va_start(val,str);
		writelog(str, val);
		va_end(val);
	}
}


static void assignthread(int fd, const char *fn, FILE *str, int at)
{
	int err;
	operator_t *op = Operators;
	pthread_attr_t attr;
	struct sched_param param;
	static unsigned numthr = 0;

	err = sem_wait(&Idle);
	assert(err == 0);
	while (op) {
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		if (op->idle) {
			/* thread went out of work */
			assert(op->fd == -1);
			assert(op->str == 0);
			assert(op->head == 0);
			assert(op->tail == 0);
			op->fd = fd;
			op->idle = 0;
			(void) strcpy(op->fn,fn);
			op->str = str;
			op->wr_err = 0;
			op->mem = 0;
			op->at = at;
			err = pthread_mutex_unlock(&op->lock);
			assert(err == 0);
			debug("assignthread(%d,\"%s\"): reusing thread %d\n",fd,fn,op->thread);
			return;
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		op = op->next;
	}
	op = calloc(1,sizeof(operator_t));
	++numthr;
	assert(numthr <= MaxThreads);
	op->at = at;
	op->fd = fd;
	(void) strcpy(op->fn,fn);
	op->str = str;
	err = pthread_mutex_init(&op->lock,0);
	assert(err == 0);
	err = pthread_cond_init(&op->oom,0);
	assert(err == 0);
	err = pthread_cond_init(&op->start,0);
	assert(err == 0);
	(void) pthread_attr_init(&attr);
	(void) pthread_attr_getschedparam(&attr,&param);
	param.sched_priority = Priority;
	(void) pthread_attr_setschedparam(&attr,&param);
	err = pthread_create(&op->thread,&attr,writer_thread,op);
	if (err != 0) {
		error("cannot create new worker thread: %s\n",strerror(errno));
		free(op);
		return;
	}
	(void) pthread_attr_destroy(&attr);
	++NumThreads;
	debug("assignthread(%d,\"%s\"): created thread %d\n",fd,fn,op->thread);
	assert(err == 0);
	op->next = Operators;
	Operators = op;
}


int open64(const char *fn, int flags, ...)
{
	int ret;
	mode_t mode = 0;

	init();
	if (flags & O_CREAT) {
		va_list val;
		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
	}
	ret = realopen64(fn,flags,mode);
	add_file(ret,fn);
	debug("open64(\"%s\",%x,%o)= %d\n",fn,flags,mode,ret);
	if (((flags & (O_RDONLY|O_WRONLY|O_RDWR)) == O_WRONLY) && (ret >= 0))
		assignthread(ret,fn,0,-1);
	return ret;
}


int open(const char *fn, int flags, ...)
{
	int ret;
	mode_t mode = 0;

	init();
	if (flags & O_CREAT) {
		va_list val;
		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
	}
	ret = realopen(fn,flags,mode);
	add_file(ret,fn);
	debug("open(\"%s\",%x,%o) = %d\n",fn,flags,mode,ret);
	if (((flags & (O_RDONLY|O_WRONLY|O_RDWR)) == O_WRONLY) && (ret >= 0))
		assignthread(ret,fn,0,-1);
	return ret;
}


FILE *fopen64(const char *fn, const char *m)
{
	FILE *ret = realfopen64(fn,m);
	if ((ret != 0) && (m[0] == 'w'))
		assignthread(fileno(ret),fn,ret,-1);
	return ret;
}


FILE *fopen(const char *fn, const char *m)
{
	FILE *ret = realfopen(fn,m);
	if ((ret != 0) && (m[0] == 'w'))
		assignthread(fileno(ret),fn,ret,-1);
	return ret;
}


int creat(const char *path, mode_t mode)
{
	int ret;
	init();
	ret = realcreat(path,mode);
	add_file(ret,path);
	debug("creat(\"%s\",%x) = %d\n",path,mode,ret);
	if (ret >= 0) 
		assignthread(ret,path,0,-1);
	return ret;
}


int creat64(const char *path, mode_t mode)
{
	int ret;
	init();
	ret = realcreat64(path,mode);
	add_file(ret,path);
	debug("creat64(\"%s\",%x) = %d\n",path,mode,ret);
	if (ret >= 0) 
		assignthread(ret,path,0,-1);
	return ret;
}


#ifdef __sun
int openat64(int at, const char *fn, int flags, ...)
{
	int ret;
	mode_t mode = 0;

	init();
	if (flags & O_CREAT) {
		va_list val;
		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
	}
	ret = realopenat64(at,fn,flags,mode);
	debug("openat64(%d,\"%s\",%x,%x) = %d\n",at,fn,flags,mode,ret);
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn,0,at);
	return ret;
}


int openat(int at, const char *fn, int flags, ...)
{
	int ret;
	mode_t mode = 0;

	init();
	if (flags & O_CREAT) {
		va_list val;
		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
	}
	ret = realopenat(at,fn,flags,mode);
	debug("openat(%d,\"%s\",%x,%x) = %d\n",at,fn,flags,mode,ret);
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn,0,at);
	return ret;
}
#endif


FILE *fdopen(int fd, const char *mode)
{
	operator_t *op = Operators;
	init();
	FILE *ret = realfdopen(fd,mode);
	while (op && (op->fd != fd)) 
		op = op->next;
	if (op && op->fd == fd) {
		int err;
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		assert(op->str == 0);
		op->str = ret;
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		debug("fdopen(%d,\"%s\") = %p -> thread %d\n",fd,mode,ret,op->thread);
	} else
		debug("fdopen(%d,\"%s\") = %p\n",fd,mode,ret);
	return ret;
}


static operator_t *add_op(op_t o, int fd, FILE *str, const char *fn)
{
	op_t *n;
	operator_t *op = Operators;
	int err;

	//debug("add_op(%d,%d,%p,\"%s\")\n",o,fd,str,fn ? fn : "<null>");
	if (fd != -1) {
		if (fn != 0) {
			while (op && (op->at != fd) && strcmp(op->fn,fn)) 
				op = op->next;
		} else while (op && (op->fd != fd)) 
			op = op->next;
	} else if (str != 0) {
		while (op && (op->str != str)) 
			op = op->next;
	} else if (fn != 0) {
		while (op && op->fn && strcmp(op->fn,fn)){
			//debug("add_op: strcmp(%s,%s)\n",op->fn,fn);
			op = op->next;
		}
	}
	if (op == 0)
		return 0;
	err = pthread_mutex_lock(&op->lock);
	assert(err == 0);
	if (op->fd == -1) {
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		return 0;
	}
	assert(op->idle == 0);
	if (op->unused) {
		n = op->unused;
		op->unused = n->next;
	} else {
		n = malloc(sizeof(op_t));
		assert(n);
	}
	*n = o;
	n->next = 0;
	if (op->head == 0) {
		op->head = op->tail = n;
	} else {
		op->tail->next = n;
		op->tail = n;
	}
	err = pthread_cond_signal(&op->start);
	assert(err == 0);
	err = pthread_mutex_unlock(&op->lock);
	assert(err == 0);
	return op;
}


int chmod(const char *path, mode_t mode)
{
	int ret = 0;
	op_t o = {do_chmod,0,mode,0,0};
	operator_t *op;
	init();
	op = add_op(o,-1,0,path);
	if (op) {
		debug("chmod(%s,%x) -> thread %d\n",path,mode,op->thread);
	} else {
		ret = realchmod(path,mode);
		debug("realchmod(%s,%x) = %d\n",path,mode,ret);
	}
	return ret;
}


int fchmod(int fd, mode_t mode)
{
	int ret = 0;
	op_t o = {do_chmod,0,mode,0,0};
	operator_t *op;
	init();
	op = add_op(o,fd,0,0);
	if (op) {
		debug("fchmod(%d,%x) -> thread %d\n",fd,mode,op->thread);
	} else {
		ret = realfchmod(fd,mode);
		debug("realfchmod(%d,%x) = %d\n",fd,mode,ret);
	}
	return ret;
}


int utime(const char *path, const struct utimbuf *times)
{
	int ret = 0;
	op_t o = {do_utime,0,0,0,0};
	operator_t *op;
	init();
	o.time.utime = *times;
	op = add_op(o,-1,0,path);
	if (op) {
		debug("utime(\"%s\",...) -> thread %d\n",path,op->thread);
	} else {
		ret = realutime(path,times);
		free(o.buf);
		debug("realutime(\"%s\",%p) = %d\n",path,times,ret);
	}
	return ret;
}


int utimes(const char *path, const struct timeval times[2])
{
	int ret = 0;
	op_t o = {do_utimes,0,0,0,0};
	operator_t *op;
	init();
	o.time.times[0] = times[0];
	o.time.times[1] = times[1];
	op = add_op(o,-1,0,path);
	if (op) {
		debug("utimes(\"%s\",...) -> thread %d\n",path,op->thread);
	} else {
		ret = realutimes(path,times);
		debug("realutimes(\"%s\",...) = %d\n",path,ret);
	}
	return ret;
}


#ifdef __sun
int futimesat(int at, const char *path, const struct timeval times[2])
{
	int ret = 0;
	const char *dn = get_name(at);
	size_t lp = strlen(path), ld = dn ? strlen(dn) : 0;
	op_t o = {do_futimesat,malloc(lp+ld+2),0,0,0};
	operator_t *op;
	init();
	o.time.times[0] = times[0];
	o.time.times[1] = times[1];
	debug("futimesat(%d [= '%s'], \"%s\", ...)\n",at,dn,path);
	(void) memcpy((char *)(o.buf),dn,ld);
	((char*)(o.buf))[ld] = '/';
	(void) memcpy((char *)(o.buf)+ld+1,path,lp+1);
	op = add_op(o,-1,0,path);
	if (op) {
		debug("futimesat(%d,\"%s\",...) -> thread %d\n",at,path,op->thread);
	} else {
		free(o.buf);
		ret = realfutimesat(at,path,times);
		debug("realfutimesat(%d,\"%s\",...) = %d\n",at,path,ret);
	}
	return ret;
}
#endif


int fclose(FILE *str)
{
	int ret = 0;
	op_t o = {do_fclose,0,0,0,0};
	operator_t *op;
	init();
	if (fileno(str) == STDERR_FILENO) {
		debug("fclose(stderr): deferred.\n");
		return 0;
	}
	op = add_op(o,-1,str,0);
	if (op) {
		debug("fclose(%p [%d]) -> thread %d\n",str,op->fd,op->thread);
	} else {
		ret = realfclose(str);
		debug("realfclose(%p) = %d\n",str,ret);
	}
	return ret;
}


int close(int fd)
{
	int ret = 0;
	op_t o = {do_close,0,0,0,0};
	operator_t *op;
	init();
	if (fd == LogFile) {
		debug("close(LogFile): deferred.\n");
		return 0;
	}
	op = add_op(o,fd,0,0);
	if (op) {
		debug("close(%d) -> thread %d\n",fd,op->thread);
	} else {
		ret = realclose(fd);
		debug("realclose(%d) = %d\n",fd,ret);
	}
	return ret;
}


int write(int f, const void *b, size_t s)
{
	static const size_t minbufsize = 1 << 20;
	int err;
	operator_t *op = Operators;
	init();
	while (op && (op->fd != f))
		op = op->next;
	if (op == 0) {
		debug("realwrite(%d,%p,%d)\n",f,b,s);
		return realwrite(f,b,s);
	}
	err = pthread_mutex_lock(&op->lock);
	assert(err == 0);
	if (op->wr_err != 0) {
		errno = op->wr_err;
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		return -1;
	}
	if (MemPerThread) {
		if (s > MemPerThread) {
			int r;
			assert(op->head == 0);
			err = pthread_mutex_unlock(&op->lock);
			assert(err == 0);
			r = realwrite(f,b,s);
			++NumDirect;
			debug("nocopy realwrite(%d,%p,%d) = %d\n",f,b,s,r);
			return r;
		}
		while (((s < MemPerThread) && (s + op->mem > MemPerThread))
				|| ((s > MemPerThread) && (op->mem > 0))) {
			err = pthread_cond_wait(&op->oom,&op->lock);
			assert(err == 0);
		}
	}
	if ((op->tail == 0)) {
		op_t *o;
		size_t bs = s < minbufsize ? minbufsize : s;
		assert(op->head == 0);
		if (op->unused) {
			o = op->unused;
			op->unused = o->next;
		} else {
			o = (op_t *) malloc(sizeof(op_t));
			assert(o);
		}
		assert(op->head == 0);
		o->op = do_write;
		o->buf = malloc(bs);
		assert(o->buf);
		(void) memcpy(o->buf,b,s);
		o->s = s;
		o->bs = bs;
		o->next = 0;
		op->head = op->tail = o;
		op->mem += bs;
		err = pthread_cond_signal(&op->start);
		assert(err == 0);
		debug("write(%d,%p,%d) -> attached to thread %d\n",f,b,s,op->thread);
	} else {
		op_t *o = op->tail;
		assert(op->head);
		assert(o->next == 0);
		size_t os = o->s;
		size_t ns = os+s;
		if (o->bs >= ns) {
			assert(o->buf);
			(void) memcpy((char *)o->buf+os,b,s);
			o->s = ns;
			debug("write(%d,%p,%d) -> thread %d, resized tail from %d to %d\n",f,b,s,op->thread,os,ns);
		} else {
			if (op->unused) {
				o->next = op->unused;
				op->unused = o->next->next;
			} else {
				o->next = (op_t *) malloc(sizeof(op_t));
				assert(o->next);
			}
			o = o->next;
			o->next = 0;
			op->tail = o;
			o->op = do_write;
			o->bs = (s < minbufsize) ? minbufsize : s;
			o->buf = malloc(o->bs);
			op->mem += o->bs;
			assert(o->buf);
			(void) memcpy(o->buf,b,s);
			o->s = s;
			err = pthread_cond_signal(&op->start);
			assert(err == 0);
			debug("write(%d,%p,%d) -> thread %d, attached to tail\n",f,b,s,op->thread);
		}
	}
	err = pthread_mutex_unlock(&op->lock);
	assert(err == 0);
	return s;
}


int fsync(int fd)
{
	operator_t *op;
	int ret = 0;
	op_t o = {do_fsync,0,0,0,0};

	op = add_op(o,fd,0,0);
	if (op) {
		debug("fsync(%d) -> thread %d\n",fd,op->thread);
	} else {
		ret = realfsync(fd);
		debug("realfsync(%d) = %d\n",fd,ret);
	}
	return ret;
}


int fflush(FILE *str)
{
	operator_t *op;
	int ret = 0;
	op_t o = {do_fflush,0,0,0,0};

	if (Abort) {
		realfflush(str);
		return 0;
	}
	op = add_op(o,-1,str,0);
	if (op) {
		debug("fflush(%p) -> thread %d\n",str,op->thread);
	} else {
		ret = realfflush(str);
		debug("realfflush(%p) = %d\n",str,ret);
	}
	return ret;
}


static void *writer_thread(void *arg)
{
	operator_t *op = (operator_t *) arg;
	op_t *unused = 0;
	pthread_t thread = op->thread;

	init();
	debug("thread %d: started\n",thread);
	for (;;) {
		op_t *o;
		int err, fd;
		char fn[PATH_MAX];
		FILE *str;

		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		if (op->head == 0) {
			err = pthread_cond_wait(&op->start,&op->lock);
			assert(err == 0);
		}
		(void) strcpy(fn,op->fn);
		fd = op->fd;
		str = op->str;
		if (unused) {
			unused->next = op->unused;
			op->unused = unused;
		}
		assert(op->head);
		o = op->head;
		if (o == op->tail) {
			assert(o->next == 0);
			op->head = op->tail = 0;
		} else {
			assert(o->next);
			op->head = o->next;
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		switch (o->op) {
		default:
			abort();
			continue;	/* to make life for lock_lint easier */

		case do_chmod:
			if (fd != -1)
				err = realfchmod(fd,o->s);
			else
				err = realchmod(fn,o->s);
			if (err == -1)
				error("error during chmod(%s,%x): %s\n",fn,o->s,strerror(errno));
			else
				debug("thread %d: realchmod(\"%s\",%x) = %d\n",thread,fn,o->s,err);
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			break;

		case do_utime:
			if (-1 == realutime(fn,&o->time.utime))
				error("error setting utime on file %s: %s\n",fn,strerror(errno));
			else
				debug("thread %d: realutime(\"%s\",...) = 0\n",thread,fn);
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			break;

		case do_utimes: 
			if (-1 == realutimes(fn,o->time.times))
				error("error setting utimes on file %s: %s\n",o->buf ? o->buf : fn, strerror(errno));
			else
				debug("thread %d: realutimes(\"%s\",...) = 0\n",thread,op->fn);
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			break;

		case do_futimesat:
			if (-1 == realutimes((char *)o->buf,o->time.times))
				error("error setting time on file %s: %s\n",o->buf,strerror(errno));
			else
				debug("thread %d: realfutimesat(\"%s\",...) = 0\n",thread,o->buf);
			free(o->buf);
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			break;

		case do_close:
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			if (realclose(fd) == -1)
				error("error closing file %s: %s\n",fn,strerror(errno));
			else
				debug("thread %d: realclose(%d [\"%s\"]) = %d\n",thread,fd,fn,err);
			op->fd = -1;
			assert(op->str == 0);
			break;

		case do_fclose:
			/* lock must be taken here, to prevent race with
			 * fdopen */
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			if (realfclose(str) != 0)
				error("error closing file %s: %s\n",fn,strerror(errno));
			else
				debug("thread %d: realfclose(%p [\"%s\"]) = 0\n",thread,str,fn);
			op->fd = -1;
			op->str = 0;
			break;

		case do_write: 
			if (op->wr_err == 0) {
				int wrerr, num = 0;
				do {
					int nw = realwrite(fd,(char *)(o->buf)+num,o->s-num);
					if (nw == -1) {
						wrerr = errno;
						break;
					}
					num += nw;
				} while (num < o->s);
				free(o->buf);
				o->buf = 0;
				err = pthread_mutex_lock(&op->lock);
				assert(err == 0);
				if (num == -1) {
					op->wr_err = wrerr;
					error("error writing to file %s: %s\n",fn,strerror(wrerr));
				} else
					debug("thread %d: realwrite(%d [\"%s\"],%p,%d) = %d\n",thread,fd,fn,o->buf,o->s,err);
			} else {
				debug("thread %d: write discarded because of previous error\n",thread);
				free(o->buf);
				o->buf = 0;
				err = pthread_mutex_lock(&op->lock);
				assert(err == 0);
			}
			op->mem -= o->bs;
			assert(op->mem >= 0);
			err = pthread_cond_signal(&op->oom);
			assert(err == 0);
			break;

		case do_fsync:
			if (-1 == realfsync(fd))
				error("thread %d: error syncing file %s: %s\n",thread,fn,strerror(errno));
			else
				debug("thread %d: fsync(%d [\"%s\"]) = 0\n",thread,fd,fn);
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			break;

		case do_fflush:
			if (EOF == realfflush(str))
				error("thread %d: error flushing file %s: %s\n",thread,fn,strerror(errno));
			else
				debug("thread %d: fflush(\"%s\") = 0\n",thread,fn);
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			break;

		case do_terminate:
			debug("thread %d: exit\n",thread);
			pthread_exit(0);
			return 0;	/* for lint */
		}
		o->next = op->unused;
		op->unused = o;
		if ((op->head == 0) && (op->fd == -1)) {
			op->idle = 1;
			assert(op->str == 0);
			assert(op->tail == 0);
			err = sem_post(&Idle);
			assert(err == 0);
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
	}
}


static void flushlog(void)
{
	while (LogHead) {
		memlog_t *lh = LogHead;
		(void) realwrite(LogFile, lh->buf, lh->at - lh->buf);
		free(lh->buf);
		LogHead = lh->next;
		free(lh);
	}
	realfsync(LogFile);
	if (LogFile == STDERR_FILENO)
		realfclose(stderr);
	realclose(LogFile);
}


static void exitproc(void)
{
	static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
	static volatile int called = 0;
	int err;

	err = pthread_mutex_lock(&mtx);
	assert(err == 0);
	if (called)
		return;
	called = 1;
	operator_t *op = Operators;
	while (op) {
		int err;
		op_t *o = (op_t *)malloc(sizeof(op_t));
		o->op = do_terminate;
		o->next = 0;
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		if (op->tail)
			op->tail->next = o;
		else
			op->head = o;
		op->tail = o;
		err = pthread_cond_signal(&op->start);
		assert(err == 0);
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		op = op->next;
	}
	op = Operators;
	while (op) {
		int err = pthread_join(op->thread,(void **)0);
		assert(err == 0);
		op = op->next;
	}
	if (Verbose) {
		logmsg("NumThreads = %d\n",NumThreads);
		logmsg("NumDirect = %d\n",NumDirect);
	}
	flushlog();
	err = pthread_mutex_unlock(&mtx);
	assert(err == 0);
}


void exit(int e)
{
	init();
	debug("exit(%d) - waiting to terminate\n",e);
	exitproc();
	realexit(e);
#ifdef __GNUC__
	for (;;);	/* just to avoid a warning */
#endif
}


void _exit(int e)
{
	init();
	debug("_exit(%d) - waiting to terminate\n",e);
	exitproc();
	real_exit(e);
#ifdef __GNUC__
	for (;;);	/* just to avoid a warning */
#endif
}


void abort(void)
{
	init();
	debug("thread %d: abort()\n",pthread_self());
	flushlog();
	Abort = 1;
	realabort();
}
